/*
 * Procedures for minimization of logic functions
 */
#include <stdio.h>
#include <math.h>
#include "logic.h"

#define FALSE	0
#define TRUE	1

/*
 * Perform generalized consensus algorithm.
 */
struct array *
consensus(f)
struct array *f;
{
	int i;
	struct array *pi, *zeroes, *ones, *exes;

	/* Change output 1's to X's */
	pi = acopy(f);
	for( i = f->a_noin; i < f->a_nvar; ++i )
		chcol(pi, i, Z, X, X);
	absorb(pi);
	for( i = 0; i < f->a_noin; ++i ){
		/* Split on column i & add intersection */
		split(pi, &zeroes, &ones, &exes, i);
		if (zeroes->a_first == NULL && ones->a_first == NULL) {
		    hass(pi, exes);
		} else {
		    hass(pi, intersect(zeroes, ones));
		    chcol(zeroes, i, Z, Z, Z);
		    append(pi, zeroes);
		    chcol(ones, i, O, O, O);
		    append(pi, ones);
		    append(pi, exes);
		    rallz(pi);
		    absorb(pi);
		    if (debug > 1) {
			printf("S%d\n", i+1);
			prary(pi, stdout);
		    }
		}
	}
	rallz(pi);
	return(pi);
}

/* Remove cubes with all-0 outputs */
rallz(pi)
register struct array *pi;
{
	struct cube *p, *p0, *q;
	int i;

	p0 = NULL;
	for( p = pi->a_first; p != NULL; p = q ){
		q = p->c_next;
		for( i = pi->a_noin; i < pi->a_nvar; ++i )
			if( p->c_logic[i] != Z )
				break;
		if( i == pi->a_nvar ){	/* all outputs zero */
			if( p0 == NULL )
				pi->a_first = q;
			else
				p0->c_next = q;
			cfree(p);
		} else
			p0 = p;
	}
	pi->a_last = p0;
}

/* Calculate no. of 0-cubes covered by an array.
   The result is a double so we can cope with > 32 inputs. */
static double
numcov(a)
struct array *a;
{
	struct array *b;
	struct cube *p;
	double cov;

	cov = 0;
	b = acopy(a);
	p = newcube(a->a_nvar);
	while( b->a_first != NULL ){
		ccopy(p, b->a_first, a->a_nvar);
		cov += ldexp(1.0, a->a_noin - cost(p, a->a_noin));
		hass(b, ashc(b, p));
	}
	afree(b);
	cfree(p);
	return cov;
}

/*
 * Returns 1 if cube q can be used to eliminate cube p, that is,
 * if cube q has an X in every glitch-free column where cube p
 * has an X.
 */
int
glcheck(glfree, p, q, noin)
    struct cube *glfree, *p, *q;
    int noin;
{
    int i;

    if (glfree == NULL)
	return 1;
    for (i = 0; i < noin; ++i)
	if (glfree->c_logic[i] && p->c_logic[i] == X && q->c_logic[i] != X)
	    return 0;
    return 1;
}

/*
 * Find essential prime implicants.
 * If none are found, the cube covering the most number of
 * 0-cubes in the on array is picked as essential.
 */
int
findepi(epi, pi, on, onleft, mininp, glfree)
struct array *epi, *pi, *on, *onleft;
int mininp;
struct cube *glfree;
{
    struct array *a, *b;
    struct cube *p, *q, *p0, **pp;
    int nused, use, pass, cos, mincos;
    struct cube **bestpp;
    double cov, maxcov;

    nused = 0;
    arr_compat(pi, on);
    p0 = NULL;
    for (pp = &pi->a_first; (p = *pp) != NULL; release(a)) {

	if (glfree == NULL) {
	    /* optimization: use the fact that onleft = on # epi. */
	    a = axc(onleft, p);

	} else {
	    /*
	     * We have some glitch-free columns; we can only eliminate
	     * p using cubes from epi or pi for which glcheck returns true.
	     */
	    a = axc(on, p);
	    for (q = epi->a_first; a->a_first != NULL && q != NULL; ){
		if (glcheck(glfree, p, q, pi->a_noin)) {
		    hass(a, ashc(a, q));
		    if (a->a_first == NULL)
			break;
		}
		q = q->c_next;
	    }
	}

	/*
	 * Discard this cube if it doesn't make any contribution
	 * not already covered by the epi's.
	 */
	if( a->a_first == NULL ){
	    /* remove p from pi */
	    *pp = p->c_next;
	    cfree(p);
	    continue;
	}

	/*
	 * See if this cube's contribution could be covered using
	 * other presently unused pi's.
	 */
	for (q = pi->a_first; q != NULL; q = q->c_next) {
	    if (q != p && (glfree == NULL
			   || glcheck(glfree, p, q, pi->a_noin))) {
		hass(a, ashc(a, q));
		if (a->a_first == NULL)
		    break;
	    }
	}

	if (a->a_first != NULL) {
	    /* this PI is essential */
	    ++nused;
	    *pp = p->c_next;
	    add(epi, p);
	    hass(onleft, ashc(onleft, p));
	    absorb(onleft);
	    if (onleft->a_first == NULL) {
		release(a);
		break;
	    }

	} else {
	    pp = &p->c_next;
	    p0 = p;
	}
    }
    pi->a_last = p0;

    if (nused > 0)
	return nused;

    /*
     * None of the pi's were essential.
     * Pick the one which makes the greatest contribution
     * to onleft for the least cost.
     */
    maxcov = 0;
    mincos = pi->a_noin;
    bestpp = NULL;
    p0 = NULL;
    for (pp = &pi->a_first; (p = *pp) != NULL; pp = &p->c_next) {
	a = axc(onleft, p);
	use = 0;
	if( mininp ){
	    cos = cost(p, pi->a_noin);
	    /* check whether this cube costs as least as much
	       as a previous cube covering at least as much. */
	    for (q = pi->a_first; q != NULL && q != p; q = q->c_next) {
		if( cos < cost(q, pi->a_noin) )
		    continue;
		b = ashc(a, q);
		if (b->a_first == NULL)
		    break;
		release(b);
	    }
	    if (q == NULL) {
		cov = numcov(a);
		if (cov * mincos > maxcov * cos) {
		    use = 1;
		    mincos = cos;
		}
	    } else {
		release(b);
	    }

	} else {
	    cov = numcov(a);
	    if (cov > maxcov)
		use = 1;
	}

	if (use) {
	    bestpp = pp;
	    maxcov = cov;
	} else
	    p0 = p;

	release(a);
    }

    if (bestpp != NULL) {
	p = *bestpp;
	if ((*bestpp = p->c_next) == NULL)
	    pi->a_last = p0;
	add(epi, p);
	hass(onleft, ashc(onleft, p));
	absorb(onleft);
	return 1;
    }

    return 0;
}

/*
 * Make ON array.
 * Each cube has only one output = 1
 */
struct array *
makeon(f)
struct array *f;
{
	struct array *on;
	struct cube *p, *new;
	int i, j;

	on = newarray(f->a_noin, f->a_nvar);
	for( p = f->a_first; p != NULL; p = p->c_next )
		for( i = f->a_noin; i < f->a_nvar; ++i )
			if( p->c_logic[i] == O ){
				new = newcube(f->a_nvar);
				for( j = 0; j < f->a_noin; ++j )
					new->c_logic[j] = p->c_logic[j];
				for( ; j < f->a_nvar; ++j )
					new->c_logic[j] = Z;
				new->c_logic[i] = O;
				add(on, new);
			}
	return(on);
}

/*
 * Perform connection minimization.
 */
conmin(e, on)
struct array *e, *on;
{
	struct cube *p0, *p, *q, *np;
	struct array *a;
	int i;

	arr_compat(e, on);
	p0 = NULL;
	for( p = e->a_first; p != NULL; release(a), p = np ){
		np = p->c_next;
		if( p0 == NULL )
			e->a_first = np;
		else
			p0->c_next = np;
		a = axc(on, p);
		hass(a, sharp(a, e));
		if( a->a_first == NULL ){
			cfree(p);
			continue;
		}
		for( i = e->a_noin; i < e->a_nvar; ++i ){
			if( p->c_logic[i] == Z )
				continue;
			p->c_logic[i] = Z;
			for( q = a->a_first; q != NULL; q = q->c_next )
				if( q->c_logic[i] == O ){
					p->c_logic[i] = O;
					break;
				}
		}
		if( p0 == NULL )
			e->a_first = p;
		else
			p0->c_next = p;
		p0 = p;
	}
	e->a_last = p0;
}

/*
 * Find minimum or near-minimum cover for switching function.
 */
struct array *
momin(f, glfree)
struct array *f;
struct cube *glfree;
{
    struct array *pi, *on, *onleft;
    struct array *epi;

    epi = newarray(f->a_noin, f->a_nvar);
    pi = consensus(f);	/* Generate prime implicants */
    on = makeon(f);
    onleft = acopy(on);
    while (onleft->a_first != NULL) {
	/* find epi's or pick one */
	if (findepi(epi, pi, on, onleft, TRUE, glfree) == 0) {
	    /* calamity! */
	    fprintf(stderr, "momin: bug: can't cover function\n");
	    break;
	}
	if (debug) {
	    printf("epi:\n");
	    prary(epi, stdout);
	    printf("onleft:\n");
	    prary(on, stdout);
	    printf("pi:\n");
	    prary(pi, stdout);
	}
    }
    conmin(epi, on);		/* Minimize connections */
    release(pi);
    release(on);
    release(onleft);
    return(epi);
}

/* Pick cubes with a given value in column i. */
struct array *
picko(f, i, u)
struct array *f;
int i, u;
{
	struct array *a;
	struct cube *p, *q;
	int j;

	a = newarray(f->a_noin, f->a_nvar);
	for( p = f->a_first; p != NULL; p = p->c_next )
		if( p->c_logic[i] == u ){
			q = newcube(f->a_nvar);
			for( j = 0; j < f->a_nvar; ++j )
				q->c_logic[j] = p->c_logic[j];
			add(a, q);
		}
	return a;
}

/* Set all outputs to zero except output i, which is changed to 1 if it's X */
seto(f, i)
struct array *f;
int i;
{
	struct cube *p;
	int j;

	for( p = f->a_first; p != NULL; p = p->c_next )
		for( j = f->a_noin; j < f->a_nvar; ++j )
			if( i != j )
				p->c_logic[j] = Z;
			else if( p->c_logic[i] == X )
				p->c_logic[i] = O;
}

/*
 * Find minimum or near-minimum cover for a switching function.
 * Each output is treated separately (as for a PAL).
 */
struct array *
somin(f, glfree)
struct array *f;
struct cube *glfree;
{
    struct array *cov, *pi, *on, *onleft, *opi, *epi;
    int i;
    
    if (debug) {
	printf("function to minimize:\n");
	prary(f, stdout);
    }
    cov = newarray(f->a_noin, f->a_nvar);
    pi = consensus(f);
    if (debug) {
	printf("pi:\n");
	prary(pi, stdout);
    }
    for (i = f->a_noin; i < f->a_nvar; ++i) {
	/* Pick terms for output i */
	on = picko(f, i, O);
	seto(on, i);
	if (debug) {
	    printf("ON for output %d:\n", i - f->a_noin + 1);
	    prary(on, stdout);
	}
	opi = picko(pi, i, X);
	epi = newarray(f->a_noin, f->a_nvar);
	onleft = acopy(on);

	/* find epi's or pick one */
	while (onleft->a_first != NULL) {
	    if (findepi(epi, opi, on, onleft, FALSE, glfree) == 0) {
		/* calamity! */
		fprintf(stderr, "somin bug: can't cover function output %d\n",
			i - f->a_noin + 1);
		break;
	    }
	    if (debug) {
		printf("epi:\n");
		prary(epi, stdout);
		printf("onleft:\n");
		prary(onleft, stdout);
		printf("remaining PI:\n");
		prary(opi, stdout);
	    }
	}

	seto(epi, i);
	append(cov, epi);
	release(opi);
	release(on);
	release(onleft);
    }
    release(pi);
    return cov;
}

